#include <oxstd.h>
#include <oxprob.h>
/**************************************************************************
 * PROGRAM:      maxelipt  -- ?th revision
 *
 * PURPOSE:
                 fit of  Elliptical Linear Mixed Models with known extra parameter 
				   
 *
 * REMARKS:      The program uses the BFGS (quasi-Newton) method of 
 *               nonlinear optimization for estimate for location and scale
 *               parameter and estimate for SE basead on Hessian Matrix and
 *               Fisher information.
 *
 *                Y_i= X_i\beta+ \epsilon_i  i =1,...,n
 *
 *                \epsilon ~ El(0,\Sigma_i)   , \Sigma_i=Z_i\tau Z_i' +\sigma*In
 *                Y_ is (px1)-vector
 *                X_i is (pxr)-matrix
 *                Z_i is (px(q-1))-matrix
 *
 * AUTHOR:       Francisco Jos de A. Cysneiros (cysneiros@de.ufpe.br).
 * 				
 *
 * LEGAL NOTE:   This program should NOT be used by or given to anyone 
 *               without explicit permission from the author. Also, the  
 *               author shall not be held liable for any damages that  
 *               possible errors or inaccuracies in the program may cause.
 *
 * VERSION:      2.0 (not ready for prime time yet, but almost...)
 *
 * DATE:         May 13, 2005.
 * 
 * LAST REVISED: .
 */
#include <oxfloat.h>
#import <maximize>
#pragma link("maximize.oxo")
static decl dg;						  // parameter of distribution
static decl fg;					 	  // parameter of distribution
static decl qsi;					  // parameter of distribution
static decl Y;                        // dependent (response) variable (p x n)
static decl XG;                       // matrix X=(X1',....Xn') matrix of covariates (p x r))
static decl ZG;				          // matrix Z=(Z1',....Zn')
static decl nuG;				      // degree of freedom
static decl kG;					      // parameter of PE distribution
static decl p;                        // dimension of explanatory variables 
static decl n;		                  // number of observations
static decl r;		                  // numbers of the location parameters 
static decl q;			              // number of restrictions of scale parameter
static const decl maxit = 10000;	  // number of iteration maximum			
static decl qd;
static decl corr;					  // corr=0, D diagonal
                                      // corr=1 , D non-diagonal
tranf(const vp)
{
decl i,j,m,taul,f;
m=0;
i=0;
j=0;
taul=zeros(qd,qd);
for(i=0;i<qd;++i){
for(j=i;j<qd;++j){
taul[i][j]=vp[m];
m=m+1;
taul[j][i]=taul[i][j];
}
}
return(taul);
}
stder(const vP){	                  //calcule Fisher information matrix
decl Phi,Sigma,vartheta,tau,i,dSigma,m,h,erps,som,B,j;
dSigma=new array[q];
Sigma=Phi=new array [n];
B=zeros(qd,qd);
erps=zeros(q,q);
if (q==1) tau=diag(0); else
{
if (corr==1) tau=tranf(vP[(r+1):(r+q-1)]);
else
tau=diag(vP[(r+1):(r+q-1)]);
}
vartheta=zeros(r,r);
	for(i = 0; i < n; ++i){
	      Sigma[i]= (ZG[i]*tau*ZG[i]')+((vP[r])*unit(p[i]));
		  Phi[i]=invertsym(Sigma[i]);
		  vartheta=vartheta+(4 .*dg[i] ./p[i]) .*(XG[i]' *Phi[i]  * XG[i]);
				}
		  vartheta =  invertsym(vartheta);
		 dSigma[]=zeros(qd,qd);
         h=1;
		 if (corr==1){ 	 
		 for(i = 0 ; i<qd ; ++i) {
		 for(j = i ; j<qd ; ++j) {
							   B[i][j] =1.0;
							   B[j][i]=B[i][j];
							   dSigma[h]=B ;
							   B=zeros(qd,qd);
							   h=h+1;
							   }
					         }	
				
}
	else   {
		for(i = 1 ; i<q ; ++i) {
							   B[(i-1)][(i-1)] =1.0;
							   dSigma[i]=B ;
						       B=zeros(qd,qd);
							   }
		}
        erps[0][0]=0.0;
		for(i=0;i<n;++i){
		 som = -0.25*(trace(Phi[i] * unit(int(p[i]))) * trace(	Phi[i] * unit(int(p[i]))))+
		  (fg[i]/(p[i]*(p[i]+2.0)))*(trace(Phi[i] * unit(int(p[i]))) * trace(Phi[i] * unit(int(p[i])))+
		  2* trace(Phi[i] * unit(int(p[i])) *Phi[i]* unit(int(p[i]))));
		  erps[0][0]=erps[0][0]+som;
		  }
	   	for( h = 1 ; h<q ; ++h){
		  erps[0][h]=0.0;
		  for(i=0;i<n;++i){
         som = -0.25*(trace(Phi[i] * unit(int(p[i]))) * trace(	Phi[i] * ZG[i]*dSigma[h]*ZG[i]'))+
		  (fg[i]/(p[i]*(p[i]+2.0)))*(trace(Phi[i] * unit(int(p[i]))) * trace(Phi[i] * ZG[i]*dSigma[h]*ZG[i]')+
		  2* trace(Phi[i] * unit(int(p[i])) * Phi[i] * ZG[i]*dSigma[h]*ZG[i]'));
		 erps[0][h]=erps[0][h]+som; 
		 }
		 erps[h][0]=erps[0][h];
		}
	for( m = 1 ; m<q ; ++m){
	   	for( h = m ; h<q ; ++h){
		  erps[m][h]=0.0;
		  for(i=0;i<n;++i){
         som = -0.25*(trace(Phi[i] * ZG[i]*dSigma[m]*ZG[i]') * trace(	Phi[i] * ZG[i]*dSigma[h]*ZG[i]'))+
		  (fg[i]/(p[i]*(p[i]+2.0)))*(trace(Phi[i] * ZG[i]*dSigma[m]*ZG[i]') * trace(Phi[i] * ZG[i]*dSigma[h]*ZG[i]')+
		  2* trace(Phi[i] * ZG[i]*dSigma[m]*ZG[i]' * Phi[i] * ZG[i]*dSigma[h]*ZG[i]'));
		 erps[m][h]=erps[m][h]+som;		
		 }
		 erps[h][m]=erps[m][h];
		} 
	}				  
	erps=invert(erps);
	return(sqrt(diagonal(vartheta))'|sqrt(diagonal(erps))');
	}
	
fLoglikN( const vP, const adFunc, const avScore,const amHessian)  // define log-likelihood function on normal error
{
decl tau,i,Phi,Sigma,di,mu,dmu,cte,extime;
Sigma=Phi=new array [n];

if (q==1) tau=diag(0); else {
if (corr==1) tau=tranf(vP[(r+1):(r+q-1)]);
else
tau=diag(vP[(r+1):(r+q-1)]);
}
cte=di = zeros(n,1);
for(i = 0; i < n; ++i){
	dmu = mu =zeros(int(p[i]),n);
	Sigma[i]= (ZG[i]*tau*ZG[i]')+((vP[r])*unit(int(p[i])));
    Phi[i]=invert(Sigma[i]);
	mu[][i] = XG[i] *vP[0:(r-1)];
	dmu[][i] = Y[i] - mu[][i];
	di[i][] = (dmu[][i])' * Phi[i] * dmu[][i];
	cte[i][]=  log(determinant(Phi[i]) );
	}
	if((vP[r]<=0) )
	{
	adFunc[0]=-DBL_EPSILON^(-1);
	   return(1);
	 }
      adFunc[0] = double(
    sumc( log(1 ./ (  M_2PI .^ ( p ./ 2.0 ))) + 0.5  * cte -0.5  .* di));//L(sigma,Y)(fdp)	;
	if (isnan(adFunc[0]) || isdotinf(adFunc[0]))
	return 0;
return 1;  // 1 indicates success
}
fLoglikT( const vP, const adFunc, const avScore,const amHessian) // define log-likelihood function on t-Student error
{
decl tau,i,Phi,Sigma,di,mu,dmu,cte,extime;
Sigma=Phi=new array [n];
if (q==1) tau=0; else {
if (corr==1) tau=tranf(vP[(r+1):(r+q-1)]);
else
tau=diag(vP[(r+1):(r+q-1)]);
}
cte=di = zeros(n,1);
	for(i = 0; i < n; ++i){
	dmu = mu =zeros(int(p[i]),n);
	Sigma[i]= (ZG[i]*tau*ZG[i]')+(vP[r]*unit(p[i]));
    Phi[i]=invert(Sigma[i]);
	mu[][i] = XG[i] *vP[0:(r-1)];
	dmu[][i] = Y[i] - mu[][i];
	di[i][] = (dmu[][i])' * Phi[i] * dmu[][i];
	cte[i]=  log(determinant(Phi[i]) );
	}
	if((vP[r]<=0) )
	{
	adFunc[0]=-DBL_EPSILON^(-1);
	   return(1);
	 }
      adFunc[0] = double(
     sumc( log(gammafact(( nuG + p ) / 2.0) ./ ( ( M_PI .* nuG ) .^ ( p ./ 2.0 ) .*
	  (gammafact( nuG ./ 2.0) ) ) )+ 0.5  *cte) +
	  sumc(-0.5 .* (nuG+p) .* log( 1.0 + di ./ nuG) ) );//L(sigma,Y)
	if (isnan(adFunc[0]) || isdotinf(adFunc[0]))
	return 0;
return 1;  // 1 indicates success
}
fLoglikPE( const vP, const adFunc, const avScore,const amHessian)	// define log-likelihood function on Power exponential error
{
decl tau,i,Phi,Sigma,di,mu,dmu,cte,extime;
Sigma=Phi=new array [n];
if (q==1) tau=0; else {
if (corr==1) tau=tranf(vP[(r+1):(r+q-1)]);
else
tau=diag(vP[(r+1):(r+q-1)]);
}
cte=di = zeros(n,1);
	for(i = 0; i < n; ++i){
	dmu = mu =zeros(int(p[i]),n);
	Sigma[i]= (ZG[i]*tau*ZG[i]')+(vP[r]*unit(p[i]));
    Phi[i]=invert(Sigma[i]);
	mu[][i] = XG[i] *vP[0:(r-1)];
	dmu[][i] = Y[i] - mu[][i];
	di[i][] = (dmu[][i])' * Phi[i] * dmu[][i];
	cte[i][]=  log(determinant(Phi[i]) );
	}
	if((vP[r]<=0) )
	{
	adFunc[0]=-DBL_EPSILON^(-1);
	   return(1);
	 }
      adFunc[0] = double(
     sumc( log(p)+loggamma(p./2)-0.5*p *log( M_PI) + 0.5  * cte-loggamma(1+(p./(2.0 ./(1+kG))))-
	 (1+(p ./(2.0 ./(1.0+kG))))*log(2) -0.5  .* (di .^(1.0 ./(1.0+kG)))));//L(sigma,Y)(fdp)	;
	if (isnan(adFunc[0]) || isdotinf(adFunc[0]))
	return 0;
return 1;  // 1 indicates success
}

main()

{
decl test,j,mx,m,vP,dfunc,avScore,ir,mhess,extime,dist,err,errI,b;
	extime = timer();		   // start the timer
	dist=0;                    // if dist=0  we have Normal model
//	dist=1;                    //if dist=1  we have t-Student model
//	dist=2;                    // if dist=2  we have PE model
	n=27;					   // number of cluster
	nuG=zeros(n,1);              //degree of freedom
	kG=zeros(n,1);	             // paramenter k in PE
	Y=new array[n];	             // response
	corr=0;						 //D diagonal
	
	p=<4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4>;	  // dimension of each cluster
	j=0;
	// data file
	XG=ZG=new array [n];
	   	mx = loadmat("c:\PROY.txt");
	// DATA MODEL
	for(m=0;m<11;++m){
	Y[m]=mx[j:j+(p[m]-1)][0] ;
	XG[m]=ones(p[m],1)~mx[j:j+(p[m]-1)][1]~zeros(p[m],1)~zeros(p[m],1);
//	ZG[m]=zeros(p[m],1);	   //  independent model
	ZG[m]=ones(p[m],1);	      // random effect intercept
//	ZG[m]=mx[j:j+(p[m]-1)][1];	      // random effect slope
//	ZG[m]=ones(p[m],1)~mx[j:j+(p[m]-1)][1];	// random effect intercept-slope
	j=j+p[m];
	nuG[m]=30;
	kG[m]=0.3;
	}
for(m=11;m<n;++m){
	Y[m]=mx[j:j+(p[m]-1)][0] ;
	XG[m]=zeros(p[m],1)~zeros(p[m],1)~ones(p[m],1)~mx[j:j+(p[m]-1)][1];
//	ZG[m]=zeros(p[m],1);	  //  independent model
	ZG[m]=ones(p[m],1);	      // random effect intercept
//	ZG[m]=mx[j:j+(p[m]-1)][1];	      // random effect slope
//	ZG[m]=ones(p[m],1)~mx[j:j+(p[m]-1)][1];	// random effect intercept-slope
    j=j+p[m];
	nuG[m]=6;
	kG[m]=0.3;
	}
if(dist==0) {
	          fg=(1.0+0.5 .*p) .*(0.5 .*p);	             // cummulant for Normal distribuition
	          dg=	p ./4.0;
			  qsi=p./p;
			  }
	else if(dist==1){		  
	fg=(nuG+p) .*(p+2) .*p .*(1 ./(4.0 .*(p+nuG+2)));	// cummalant for student-t distribution	
	dg=	(p.*(nuG+p))./(4 .*(nuG+p+2));
	qsi=nuG./(nuG-2);
	}
	else if(dist==2){
    fg= (1 ./((1+kG) .^2)) .*( p .*(1+kG) ./ 2.0 +1).*( p .*(1+kG) ./ 2.0); // cummalant for PE distribution
    dg=((1 ./(((1+kG).^(2.0)) .*( 0.5 .^(kG+1)))) .* ( gammafact(((p .* (1+kG))-2 .*kG+2).* 0.5)))./
	( gammafact(p .* (1+kG) .* 0.5) ) ;
	qsi=2.0^(1+kG).*(gammafact(0.5*(p+2).*(kG+1))) ./ (p.*gammafact(0.5.*p.*(kG+1)));
	}
// model matrix					
	r=columns(XG[0]);
	q=columns(ZG[0])+1; // component variance model
	qd=q-1; 
//  q=qd=1; 	       // independent model
    if (corr==1) {
	qd=columns(ZG[0]);
	q=0.5*(qd^2-qd)+qd+1;
	}	
    errI=zeros(q,1)';  
	mhess=unit(r+q);                  // initialize the matrixHessian/*
	vP=ones(r+q,1);                   //initial value
	MaxControlEps( 1e-6, 1e-4);       // set the accuracy for BFGS estimation
	    MaxControl(10000000,-1,-1);   // set maximum # of iterations = 1000
		if(dist==0){
        ir = MaxBFGS(fLoglikN, &vP, &dfunc, 1,1);
		if(Num2Derivative(fLoglikN,vP,&mhess)){err=sqrt(diagonal(-invert(mhess)));
	    }
		}
		else
		if(dist==1){ir = MaxBFGS(fLoglikT, &vP, &dfunc,&mhess,1);
		if(Num2Derivative(fLoglikT,vP,&mhess)){err=sqrt(diagonal(-invert(mhess)));}
		}
		else if(dist==2){ir = MaxBFGS(fLoglikPE, &vP, &dfunc, &mhess,1);
		if(Num2Derivative(fLoglikPE,vP,&mhess)){err=sqrt(diagonal(-invert(mhess)));}
		}
		else {println( " ERROR- absent distribution");}
		if (ir!= MAX_CONV && ir != MAX_WEAK_CONV) test=0; else test=1; //fracass

if(q!=1)
{
        decl Gmu,GSigma,GPhi,Gpred;
		decl res,vaux,waux,raux,saux,rnv,renv,v,u,tau,i,w,l,mu;
		decl GY,SigmaG,GX,GZ,GD,Gb,GA,renvc;
		decl fitt;
saux=GSigma=GPhi=new array [n];
v=u=zeros(n,1);
GY=zeros(int(sumc(p)),1);
SigmaG=zeros(int(sumc(p)),int(sumc(p)));
GX=zeros(int(sumc(p)),r);
GZ=zeros(int(sumc(p)),n*qd);
j=l=w=0;
res=fitt=zeros(sumc(p),1);
if (q==1) tau=0; else {
if (corr==1) tau=tranf(vP[(r+1):(r+q-1)]);
else
tau=diag(vP[(r+1):(r+q-1)]);
}
	for(i = 0; i < n; ++i){
    Gpred=Gmu =zeros(int(p[i]),n);
	GSigma[i]= (ZG[i]*tau*ZG[i]')+(vP[r]*unit(p[i]));
    GPhi[i]=invert(GSigma[i]);
	Gmu[][i] = XG[i] *vP[0:(r-1)];
	Gpred[][i] = (Y[i] - Gmu[][i]);
	u[i]=Gpred[][i]'*GPhi[i]*Gpred[][i];
		if(dist==0) { v[i]=1;}				//for  Normal
	else if(dist==1){ v[i]=(nuG[i]+p[i])./(nuG[i]+u[i]);}//	for student-t
	else if(dist==2){v[i]=1.0./((1+kG[i]).*(u[i].^(kG[i]./(kG[i]+1))));} // for EP
saux[i] = zeros(int(p[i]),int(p[i]));
vaux=new array[1];
waux=new array[1];
raux=new array[1];
 //fitted value
	 	 decsvd(GSigma[i],vaux,waux,raux);
         saux[i] = ((vaux[0] * (raux[0]' .*(1.0 ./sqrt( waux[0]))') )') ;
	     res[j:j+(p[i]-1)]=saux[i] *(Gpred[][i]).*sqrt(v[i]);
		 GY[j:j+(p[i]-1)][]=Y[i];
		 SigmaG[j:j+(p[i]-1)][j:j+(p[i]-1)]=GSigma[i];
		 GX[j:j+(p[i]-1)][]=XG[i];
		 GZ[j:j+(p[i]-1)][w:w+(qd-1)]=ZG[i];
		 j=j+p[i];
		 l=l+r;
		 w=w+qd;
	 }	 
	GD=unit(int(n))**tau;  
	Gb=GD*GZ'*invert(SigmaG)*(GY-GX*vP[0:(r-1)]);
	fitt=(GX*vP[0:(r-1)])-GZ*Gb;
	GA=unit(int(sumc(p)))-GZ*GD*GZ'*invert(SigmaG);
	renvc=GA*(GY-GX*vP[0:(r-1)]);
	savemat("m1.txt", fitt);  //fitted values
	savemat("m2.txt",renvc); //conditional residuals
}
	print( "\n\t\t Potthoff-Roy ");
    print( "\n\t\t ------------------\n");
    print( "\n\t\t OX PROGRAM: ", oxfilename(0) );
    print( "\n\t\t OX VERSION: ", oxversion() );
	if(dist==0){
	print( "\n\t\t DISTRIBUTION: Normal ");}
	else 	if(dist==1){
	print( "\n\t\t DISTRIBUTION: T-student D.F Known ");}
    else 	if(dist==2){
	print( "\n\t\t DISTRIBUTION: Power Exponential");}
	else print (" not implemenent yet");
	if (q==1) print( "\n\t\t structured variance : independent");
	else print( "\n\t\t structured variance : component variance");
	print( "\n\t\t NUM. PARAMETERS IN COV: ", q );
	print( "\n\t\t NUM. PARAMETERS IN MODEL: ", r );										   
	if(corr==1) print("\n\t\t  D : general structure") ; else	 print("\n\t\t D : Diagonal");
	if (any(p .!= p[0]))  print( "\n\t\t NUM. DIMENSION RESPONSE: unbalanced");
	else print ( "\n\t\t NUM. DIMENSION RESPONSE: balanced p=" ,p[0]);
    print( "\n\t\t NUM. OBSERVATIONS: ", n );
    print( "\n\t\t DATE: ", date() );
    print( "\n\t\t TIME: ", time() );
	print( "\nEXECUTION TIME: ", timespan(extime) , "\n");
		if(test==1) {
	println("\n",MaxConvergenceMsg(ir),"\nFunction value = ", dfunc);
	println("\n",MaxConvergenceMsg(ir),"\nAkaike = ", -2*dfunc + 2*(r+q));
	errI=stder(vP)'
		println( "\n PARAMETER ESTIMATES:" );
		print( "%r", {"Estimate param by ML:","EP:","EP(Fisher)"},
	       "%c",  {"bo","b1", "b2","b3","sigma^2","tau's"},
		       "%10.4f", vP'|err |errI);   
	}
	else{println(" NO CONVERGE");
		println("\n",MaxConvergenceMsg(ir));
	}
	println("---------------------------------------------");						 			
}			