#include <oxstd.h>
#include <oxprob.h>
#import  <solveqp>
/**************************************************************************
 * PROGRAM:      TS.ox  
 *
 * PURPOSE:		  evaluate the calculation of the TS statistics for the situations 
 *				  mentioned in the paper by assuming elliptical distribution for the errors
 *
 * AUTHOR:      Francisco Jos de A. Cysneiros (cysneiros@de.ufpe.br)
 *				
 * 				
 *
 * LEGAL NOTE:   This program should NOT be used by or given to anyone 
 *               without explicit permission from the author. Also, the  
 *               author shall not be held liable for any damages that  
 *               possible errors or inaccuracies in the program may cause.
 *
 * VERSION:      1.0 (not ready for prime time yet, but almost...)
 *
 * DATE:         May 9, 2005.
 * 
 * LAST REVISED: 	 May 9, 2005
 */
#include <oxfloat.h>
#import <maximize>
#pragma link("maximize.oxo")
static decl dg;						  // parameter of distribution
static decl fg;					 	  // parameter of distribution
static decl qsi;					  // parameter of distribution
static decl Y;                        // dependent (response) variable (p x n)
static decl XG;                       // matrix X=(X1',....Xn') matrix of covariates (p x r))
static decl ZG;				          // matrix Z=(Z1',....Zn')
static decl nuG;				      // degree of freedom
static decl kG;					      // parameter of PE distribution
static decl p;                        // dimension of explanatory variables 
static decl n;		                  // number of observations
static decl r;		                  // numbers of the location parameters 
static decl q;			              // number of restrictions of scale parameter
static const decl maxit = 10000;	  // number of iteration maximum			
static decl dist;                     // distribution
static decl corr;					  // corr=0 , D diagonal
                                      //, corr=1 , D non-diagonal
static decl qd;
 //transform on matrix D 
tranf(const vp)
{
decl i,j,m,taul,f;
m=0;
i=0;
j=0;
taul=zeros(qd,qd);
for(i=0;i<qd;++i){
for(j=i;j<qd;++j){
taul[i][j]=vp[m];
m=m+1;
taul[j][i]=taul[i][j];
}
}
return(taul);
}
/*  Score function */
fescore( const vP){
decl tau,i,j,Phi,Sigma,u,mu,dmu,Gmu,cte,v,dSigma,Sb,Sg,Stau,som,h,B;
Sigma=Phi=mu=Gmu=new array [n];
dSigma=new array [q];
B=zeros(qd,qd);
v=zeros(n,1);
Sb=zeros(r,n);
Sg=zeros(1,n);
Stau=zeros(q,1);
if (q==1) tau=diag(0); else
{
if (corr==1) tau=tranf(vP[(r+1):(r+q-1)]);
else
tau=diag(vP[(r+1):(r+q-1)]);
}
    u = zeros(n,1);
    dSigma[]=zeros(q-1,q-1);
	for(i = 0; i < n; ++i){
	Sigma[i]= (ZG[i]*tau*ZG[i]')+((vP[r])*unit(int(p[i])));
    Phi[i]=invert(Sigma[i]);
	mu[i] = XG[i] *vP[0:(r-1)];
	Gmu[i] = Y[i] - mu[i];
	u[i][] = (Gmu[i])' * Phi[i] * Gmu[i];
	if(dist==0) { v[i]=1;}				//for  Normal
	else if(dist==1){ v[i]=(nuG[i]+p[i])./(nuG[i]+u[i]);}//	for student-t
	else if(dist==2){v[i]=1.0./((1+kG[i]).*(u[i].^(kG[i]./(kG[i]+1))));} // for EP
	Sb[][i]=v[i].*XG[i]'*Phi[i]*mu[i];
	Sg[0][i] = -0.5*(trace(Phi[i] * unit(int(p[i])))-v[i].*Gmu[i]'*Phi[i]*unit(int(p[i]))*Phi[i]*Gmu[i]);
	}
    dSigma[]=zeros(qd,qd);
	h=1;
	 if (corr==1){ 	 
		 for(i = 0 ; i<qd ; ++i) {
		 for(j = i ; j<qd ; ++j) {
							   B[i][j] =1.0;
							   B[j][i]=B[i][j];
							   dSigma[h]=B ;
							   B=zeros(qd,qd);
							   h=h+1;
							   }
					         }	
				
}
	else   {
		for(i = 1 ; i<q ; ++i) {
							   B[(i-1)][(i-1)] =1.0;
							   dSigma[i]=B ;
						       B=zeros(qd,qd);
							   }
		}	
	for(h = 1 ; h<q ; ++h){
	    Stau[h][0]=0.0;
		for(i=0;i<n;++i){
        som =  -0.5.*(trace(Phi[i]*ZG[i]*dSigma[h]*ZG[i]') -v[i].*Gmu[i]'*Phi[i]*ZG[i]*dSigma[h]*ZG[i]'*Phi[i]*Gmu[i]);
		Stau[h][0]=Stau[h][0]+som;
		}
	}
	Stau[0][0]=sumr(Sg);
	return (sumr(Sb)|Stau);
}

/*    Fisher Information Matrix	 */
Fisher(const vP){	                  
decl Phi,Sigma,vartheta,tau,i,j,dSigma,m,h,erps,som,B,K;
K=zeros((r+q),(r+q));
dSigma=new array[q];
Sigma=Phi=new array [n];
B=zeros(qd,qd);
erps=zeros(q,q);
if (q==1) tau=diag(0); else
{
if (corr==1) tau=tranf(vP[(r+1):(r+q-1)]);
else
tau=diag(vP[(r+1):(r+q-1)]);
}
vartheta=zeros(r,r);
	for(i = 0; i < n; ++i){
	      Sigma[i]= (ZG[i]*tau*ZG[i]')+((vP[r])*unit(p[i]));
		  Phi[i]=invertsym(Sigma[i]);
		  vartheta=vartheta+(4 .*dg[i] ./p[i]) .*(XG[i]' *Phi[i]  * XG[i]);
				}
		  vartheta =  invertsym(vartheta);
		 dSigma[]=zeros(qd,qd);
         h=1;
		 if (corr==1){ 	 
		 for(i = 0 ; i<qd ; ++i) {
		 for(j = i ; j<qd ; ++j) {
							   B[i][j] =1.0;
							   B[j][i]=B[i][j];
							   dSigma[h]=B ;
							   B=zeros(qd,qd);
							   h=h+1;
							   }
					         }	
				
}
	else   {
		for(i = 1 ; i<q ; ++i) {
							   B[(i-1)][(i-1)] =1.0;
							   dSigma[i]=B ;
						       B=zeros(qd,qd);
							   }
		}
        erps[0][0]=0.0;
		for(i=0;i<n;++i){
		 som = -0.25*(trace(Phi[i] * unit(int(p[i]))) * trace(	Phi[i] * unit(int(p[i]))))+
		  (fg[i]/(p[i]*(p[i]+2.0)))*(trace(Phi[i] * unit(int(p[i]))) * trace(Phi[i] * unit(int(p[i])))+
		  2* trace(Phi[i] * unit(int(p[i])) *Phi[i]* unit(int(p[i]))));
		  erps[0][0]=erps[0][0]+som;
		  }
	   	for( h = 1 ; h<q ; ++h){
		  erps[0][h]=0.0;
		  for(i=0;i<n;++i){
         som = -0.25*(trace(Phi[i] * unit(int(p[i]))) * trace(	Phi[i] * ZG[i]*dSigma[h]*ZG[i]'))+
		  (fg[i]/(p[i]*(p[i]+2.0)))*(trace(Phi[i] * unit(int(p[i]))) * trace(Phi[i] * ZG[i]*dSigma[h]*ZG[i]')+
		  2* trace(Phi[i] * unit(int(p[i])) * Phi[i] * ZG[i]*dSigma[h]*ZG[i]'));
		 erps[0][h]=erps[0][h]+som; 
		 }
		 erps[h][0]=erps[0][h];
		}
	for( m = 1 ; m<q ; ++m){
	   	for( h = m ; h<q ; ++h){
		  erps[m][h]=0.0;
		  for(i=0;i<n;++i){
         som = -0.25*(trace(Phi[i] * ZG[i]*dSigma[m]*ZG[i]') * trace(	Phi[i] * ZG[i]*dSigma[h]*ZG[i]'))+
		  (fg[i]/(p[i]*(p[i]+2.0)))*(trace(Phi[i] * ZG[i]*dSigma[m]*ZG[i]') * trace(Phi[i] * ZG[i]*dSigma[h]*ZG[i]')+
		  2* trace(Phi[i] * ZG[i]*dSigma[m]*ZG[i]' * Phi[i] * ZG[i]*dSigma[h]*ZG[i]'));
		 erps[m][h]=erps[m][h]+som;		
		 }
		 erps[h][m]=erps[m][h];
		} 
	}	
	K[0:(r-1)][0:(r-1)]=vartheta;
	K[r:(r+q-1)][r:(r+q-1)]=erps;
	return(K);
	}
/* ...........................Function of the Partial correlation........................... */
parc(const aCor,const ai, const aj, const ak){
decl St;
  St = (aCor[ai][aj] - aCor[ai][ak] * aCor[aj][ak])/
  sqrt((1 - aCor[ai][ak]^2) * (1 - aCor[aj][ak]^2));
return St;
}

/* ....................weights - orthants probabilities................... */
/* Wolak (1986) - JASA
*/
pesos( const KT){
decl cor,pes,B,ak,i,j,pl;
    ak=columns(KT);
	pes = zeros(ak+1,1);
	cor = zeros(ak,ak);
	if (ak==1) {
	       pes[0]=0.5;
		   pes[1]=0.5;
    }
	else {
		B = KT;
		for(i = 0; i < ak; ++i){
			for(j = 0; j < ak; ++j){
				cor[i][j] = B[i][j]/sqrt(B[i][i] * B[j][j]);
			}
		}
	if (ak==2) {
		pes[0] = acos(cor[0][1])/M_2PI;
		pes[1] =0.5; 
		pes[2] = 0.5-pes[0];  
 	}
	else if (ak==3) {
		pl[0] =  parc(cor,0,1,2);
		pl[1] =  parc(cor,0,2,1);
		pl[2] =  parc(cor,1,2,0);
		pes[2] = (3.0 * M_PI - acos(pl[0]) - acos(pl[1]) - acos(pl[2]))/(4*M_PI);
		pes[3] = (2.0 * M_PI - acos(cor[0][1]) - acos(cor[0][2]) - acos(cor[1][2]))/(4*M_PI);
		pes[1] = 0.5 - pes[3];
		pes[0] = 0.5 - pes[2];
 	}
 }
return pes;
}	
main()

{
decl p1,S,St,K,Kts,KI,KIss,Ss,KItt,Z,TS;
decl iret,mg,vg,ma,vbs,x,TS1,Q,pTS1,k,d;
decl q0,i,j,mx,m,vP,extime,vP0;
	extime = timer();		   // start the timer
    dist=0;                    // if dist=0  we have Normal model
//	dist=1;                    // if dist=1  we have t-Student model
	dist=2;                    // if dist=2  we have PE model
	n=20;	                   // number of clusters
	nuG=zeros(n,1);            // degree of freedom	of t-model
	kG=zeros(n,1);	           // paramenter k in PE model
	qsi=zeros(n,1);			   // parameter qsi
	Y=new array[n];	           // response variable
	corr=1;						 //random effect correlated
	p=<4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4;4>;
	j=0;
	// data file
	XG=ZG=new array [n];
	   	mx = loadmat("c:\RAMUS_ox.txt");
	// DATA MODEL
	for(m=0;m<20;++m){
	Y[m]=mx[j:j+(p[m]-1)][2] ;
	XG[m]=ones(p[m],1)~mx[j:j+(p[m]-1)][1];
	ZG[m]=ones(p[m],1);	                      // random effect intercept
//	ZG[m]=mx[j:j+(p[m]-1)][1];	              // random effect slope
//	ZG[m]=ones(p[m],1)~mx[j:j+(p[m]-1)][1]; 	// random effect intercept-slope
	j=j+p[m];
	nuG[m]=30;
	kG[m]=0.6;
}
	r=columns(XG[0]);
	q=columns(ZG[0])+1;               // component variance model
	qd=q-1; 
	if (corr==1) {
	qd=columns(ZG[0]);
	q=0.5*(qd^2-qd)+qd+1;
	}
	// Normal-Model
if(dist==0){
//	vP0=<33.75;1.860;6.4714>;	// estimates under Ho:  Ho:tau_1=0 x tau_1>0
//	vP0=<33.75;1.860;6.4714>;	// estimates under Ho:  Ho:tau_2=0 x tau_2>0
//	vP0=<33.75;1.860;6.4714>;	// estimates under Ho:  Ho:tau1=tau_2=0 x at least one >0
	vP0=<34.0487;1.8367;0.618;5.907>; // estimates under Ho:  Ho:tau_2=0 , tau_{12}=0 x tau_1>0  | tau_1, tau_{12} present 
}			  
 //PE-model
if(dist==2){
//	vP0=<33.981;1.8269;1.1652>;   // estimates under Ho:  Ho:tau_1=0 x tau_1>0
//  vP0=<33.981;1.8269;1.1652>; 	// estimates under Ho:  Ho:tau_2=0 x tau_2>0
//	vP0=<33.981;1.8269;1.1652>; 	// estimates under Ho:  Ho:tau1=tau_2=0 x at least one >0
	vP0=<35.8866;1.5861;0.0926;1.4675>; // estimates under Ho:  Ho:tau_2=0 , tau_{12}=0 x tau_2>0  | tau_1, tau_{12} present 
}					 
	if(dist==0) {
	          fg=(1.0+0.5 .*p) .*(0.5 .*p);	             // cummulant for Normal distribuition
	          dg=	p ./4.0;
			  qsi=p./p;
			  }
	else if(dist==1){		  
	fg=(nuG+p) .*(p+2) .*p .*(1 ./(4.0 .*(p+nuG+2)));	// cummalant for student-t distribution	
	dg=	(p.*(nuG+p))./(4 .*(nuG+p+2));
	qsi=nuG./(nuG-2);
	}
	else if(dist==2){
    fg= (1 ./((1+kG) .^2)) .*( p .*(1+kG) ./ 2.0 +1).*( p .*(1+kG) ./ 2.0); // cummalant for PE distribution
    dg=((1 ./(((1+kG).^(2.0)) .*( 0.5 .^(kG+1)))) .* ( gammafact(((p .* (1+kG))-2 .*kG+2).* 0.5)))./
	( gammafact(p .* (1+kG) .* 0.5) ) ;
	qsi=2.0^(1+kG).*(gammafact(0.5*(p+2).*(kG+1))) ./ (p.*gammafact(0.5.*p.*(kG+1)));
	}
	q0=rows(vP0);
	vP=zeros(r+q,1);   
	vP[0:(q0-1)][]=vP0;
    S=fescore(vP);
    K=Fisher(vP);
    KI=invert(K);
    St=S[(q0):(r+q-1)][];
    Kts=K[(q0):(r+q-1)][r]	;
    KIss=invert(K[r][r]);
    Ss=S[r][];
    Z=St-Kts*KIss*Ss;
    KItt=KI[(q0):(r+q-1)][(q0):(r+q-1)];
    TS=Z'*KItt*Z;
    mg = 2*KItt;
    vg = -2*KItt*Z;
    if(corr==1)
	{
	ma = unit(columns(KItt));
	vbs = zeros(columns(KItt),1);
	}
	else
	{
	ma=<0,1>;
	vbs=<0>;
	}
    [iret,x] = SolveQP(mg, vg, ma, vbs, <>, <>, <>, <>);
	// TS TEST
	TS1=-(x'mg*x)/2 - vg'x;
    k=columns(KItt);
    pTS1=0.0;
	Q=zeros(k+1,1);
	for(m=0;m<=k;++m){	   //approximate weights
	Q[m][0]=binomial(k,m)*(2.0^(-k));}
	Q=pesos(KItt);	  // Weights (shapiro)
	if(corr==1){
	//calculate  pvalue - 	TS1 
	if(TS1 == 0.0) {pTS1 = Q[0][0];}
	else {		 
	    pTS1=0.5*(1-probchi(TS1, 1))+0.5*(1-probchi(TS1, 2));
		}
	} else
	//calculate  pvalue - 	TS1 
	if(TS1 == 0.0) {pTS1 = Q[0][0];}
	else {		 
		for(m=1;m<=k;++m){
	    pTS1 = pTS1 + (Q[m][0]*(1-probchi(TS1, m)));}
		}
	print( "\n\t\t ------------------\n");
    print( "\n\t\t OX PROGRAM: ", oxfilename(0) );
    print( "\n\t\t OX VERSION: ", oxversion() );
	if(dist==0){
	print( "\n\t\t DISTRIBUTION: Normal ");}
	else 	if(dist==1){
	print( "\n\t\t DISTRIBUTION: T-student D.F Known ");}
    else 	if(dist==2){
	print( "\n\t\t DISTRIBUTION: Power Exponential");}
	else print (" not implemenent yet");
	print( "\n\t\t H0: tau's=0   x  tau's > 0");
	print( "\n\t\t NUM. PARAMETERS IN H0: ", q0 );
	print( "\n\t\t NUM. PARAMETERS IN H1: ", q+r );
	if(corr==1) print("\n\t\t  D : general structure") ; else	 print("\n\t\t D : Diagonal");
	if (any(p .!= p[0]))  print( "\n\t\t NUM. DIMENSION RESPONSE: unbalanced");
	else print ( "\n\t\t NUM. DIMENSION RESPONSE: balanced p=" ,p[0]);
    print( "\n\t\t NUM. OBSERVATIONS: ", n );
    print( "\n\t\t DATE: ", date() );
    print( "\n\t\t TIME: ", time() );
	print( "\n\t\t EXECUTION TIME: ", timespan(extime) , "\n");
	println("");
	println( "\n PARAMETER ESTIMATES UNDER H0:" );
	print( "%r", {"Estimate param by ML:"},
	       "%c",  {"a", "b","sigma^2","tau's"},
		       "%10.4f", vP');	   
    print("\n\t\tThe value of Score test  = ", double(TS1));
    println(" and p-value = ", double(pTS1));

}